'''
    icmplib
    ~~~~~~~

        https://github.com/ValentinBELYN/icmplib

    :copyright: Copyright 2017-2020 Valentin BELYN.
    :license: GNU LGPLv3, see the LICENSE for details.

    ~~~~~~~

    This program is free software: you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public License
    as published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this program.  If not, see
    <https://www.gnu.org/licenses/>.
'''

from os import getpid
from random import choice


PID = getpid()


def random_byte_message(size):
    '''
    Generate a random byte sequence of the specified size.

    '''
    available_bytes = (
        b'abcdefghijklmnopqrstuvwxyz'
        b'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
        b'1234567890'
    )

    sequence = [
        choice(available_bytes)
        for _ in range(size)
    ]

    return bytes(sequence)


def is_ipv6_address(address):
    '''
    Take an IP address and indicate whether it is an IPv6 address or
    not. Return a boolean.

    '''
    return ':' in address



'''
    icmplib
    ~~~~~~~

        https://github.com/ValentinBELYN/icmplib

    :copyright: Copyright 2017-2020 Valentin BELYN.
    :license: GNU LGPLv3, see the LICENSE for details.

    ~~~~~~~

    This program is free software: you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public License
    as published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this program.  If not, see
    <https://www.gnu.org/licenses/>.
'''

from .ip import IPv4Socket, IPv6Socket
from .models import ICMPRequest, ICMPReply
from .exceptions import *
from .utils import random_byte_message

import socket
from struct import pack, unpack
from time import time


#   Echo Request and Echo Reply messages -- RFC 792 / 4443
#
#      0                   1                   2                   3
#      0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     |     Type      |     Code      |           Checksum            |
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     |           Identifier          |        Sequence Number        |
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     |     Data ...
#     +-+-+-+-+-
#
#   ICMPv4 Error message -- RFC 792
#
#      0                   1                   2                   3
#      0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     |     Type      |     Code      |           Checksum            |
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     |                 Unused / Depends on the error                 |
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     |      Internet Header + 64 bits of Original Data Datagram      |
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#
#   ICMPv6 Error message -- RFC 4443
#
#      0                   1                   2                   3
#      0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     |     Type      |     Code      |           Checksum            |
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     |                 Unused / Depends on the error                 |
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
#     |      Original packet without exceed the minimum IPv6 MTU      |
#     +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+


class ICMPConfig:
    IP_PROTOCOL          = -1
    IP_SOCKET            = None

    ICMP_HEADER_OFFSET   = -1
    ICMP_CODE_OFFSET     = ICMP_HEADER_OFFSET + 1
    ICMP_CHECKSUM_OFFSET = ICMP_HEADER_OFFSET + 2
    ICMP_ID_OFFSET       = ICMP_HEADER_OFFSET + 4
    ICMP_SEQUENCE_OFFSET = ICMP_HEADER_OFFSET + 6
    ICMP_PAYLOAD_OFFSET  = ICMP_HEADER_OFFSET + 8

    ICMP_ECHO_REQUEST    = -1
    ICMP_ECHO_REPLY      = -1


class ICMPv4Config(ICMPConfig):
    IP_PROTOCOL          = 1
    IP_SOCKET            = IPv4Socket

    ICMP_HEADER_OFFSET   = 20
    ICMP_CODE_OFFSET     = ICMP_HEADER_OFFSET + 1
    ICMP_CHECKSUM_OFFSET = ICMP_HEADER_OFFSET + 2
    ICMP_ID_OFFSET       = ICMP_HEADER_OFFSET + 4
    ICMP_SEQUENCE_OFFSET = ICMP_HEADER_OFFSET + 6
    ICMP_PAYLOAD_OFFSET  = ICMP_HEADER_OFFSET + 8

    ICMP_ECHO_REQUEST    = 8
    ICMP_ECHO_REPLY      = 0


class ICMPv6Config(ICMPConfig):
    IP_PROTOCOL          = 58
    IP_SOCKET            = IPv6Socket

    ICMP_HEADER_OFFSET   = 0
    ICMP_CODE_OFFSET     = ICMP_HEADER_OFFSET + 1
    ICMP_CHECKSUM_OFFSET = ICMP_HEADER_OFFSET + 2
    ICMP_ID_OFFSET       = ICMP_HEADER_OFFSET + 4
    ICMP_SEQUENCE_OFFSET = ICMP_HEADER_OFFSET + 6
    ICMP_PAYLOAD_OFFSET  = ICMP_HEADER_OFFSET + 8

    ICMP_ECHO_REQUEST    = 128
    ICMP_ECHO_REPLY      = 129


class ICMPSocket:
    '''
    Base class for ICMP sockets.

    :type config: ICMPConfig
    :param config: The ICMP socket configuration used to create and
        read ICMP packets.

    :raises SocketPermissionError: If the permissions are insufficient
        to create the socket.

    '''
    def __init__(self, config):
        self._socket = None
        self._config = config
        self._last_request = None

        try:
            self._socket = config.IP_SOCKET(
                config.IP_PROTOCOL)

        except OSError:
            raise SocketPermissionError

    def __del__(self):
        '''
        Call the close() method.

        '''
        self.close()

    def _create_header(self, type, code, checksum, id, sequence):
        '''
        Create the ICMP header of a packet.

        '''
        #  8 bits: B
        # 16 bits: H
        return pack('!2B3H', type, code, checksum, id, sequence)

    def _checksum(self, data):
        '''
        Calculate the checksum of a packet.

        '''
        end = len(data) - 1
        sum = 0

        for i in range(0, end, 2):
            sum += (data[i] << 8) + data[i + 1]
            sum  = (sum >> 16) + (sum & 0xffff)

        sum = ~sum & 0xffff

        return sum

    def _create_packet(self, id, sequence, payload):
        '''
        Create a packet.

        '''
        type = self._config.ICMP_ECHO_REQUEST

        # Temporary header to calculate the checksum
        header = self._create_header(type, 0, 0, id, sequence)
        checksum = self._checksum(header + payload)

        # Definitive header
        header = self._create_header(type, 0, checksum, id, sequence)

        return header + payload

    def _read_reply(self, packet, source, reply_time):
        '''
        Read a reply from bytes and return an ICMPReply object.

        '''
        received_bytes = (
            len(packet)
            - self._config.ICMP_HEADER_OFFSET)

        type, code = unpack('!2B', packet[
            self._config.ICMP_HEADER_OFFSET:
            self._config.ICMP_CHECKSUM_OFFSET])

        if type != self._config.ICMP_ECHO_REPLY:
            packet = packet[self._config.ICMP_PAYLOAD_OFFSET:]

        id, sequence = unpack('!2H', packet[
            self._config.ICMP_ID_OFFSET:
            self._config.ICMP_PAYLOAD_OFFSET
        ])

        reply = ICMPReply(
            source=source,
            id=id,
            sequence=sequence,
            type=type,
            code=code,
            received_bytes=received_bytes,
            time=reply_time)

        return reply

    def send(self, request):
        '''
        Send a request to a host.

        This operation is non-blocking. Use the receive() method to get
        the reply.

        :type request: ICMPRequest
        :param request: The ICMP request you have created.

        :raises SocketBroadcastError: If a broadcast address is used
            and the corresponding option is not enabled on the socket
            (ICMPv4 only).
        :raises ICMPSocketError: If another error occurs while sending.

        '''
        payload = request.payload

        if not payload:
            payload = random_byte_message(
                size=request.payload_size)

        packet = self._create_packet(
            id=request.id,
            sequence=request.sequence,
            payload=payload)

        self._socket.ttl = request.ttl

        request._time = time()

        try:
            self._socket.send(
                packet=packet,
                address=request.destination,
                port=0)

            self._last_request = request

        except PermissionError:
            raise SocketBroadcastError

        except OSError as err:
            raise ICMPSocketError(str(err))

    def receive(self):
        '''
        Receive a reply from a host.

        This method can be called multiple times if you expect several
        responses (as with a broadcast address).

        :raises TimeoutExceeded: If no response is received before the
            timeout defined in the request.
            This exception is also useful for stopping a possible loop
            in case of multiple responses.
        :raises ICMPSocketError: If another error occurs while
            receiving.

        :rtype: ICMPReply
        :returns: An ICMPReply object containing the reply of the
            desired destination.

        See the ICMPReply class for details.

        '''
        if not self._last_request:
            raise TimeoutExceeded(0)

        request = self._last_request

        current_time = time()
        self._socket.timeout = request.timeout
        timeout = current_time + request.timeout

        try:
            while True:
                packet, address, port = self._socket.receive()
                reply_time = time()

                if reply_time > timeout:
                    raise socket.timeout

                reply = self._read_reply(
                    packet=packet,
                    source=address,
                    reply_time=reply_time)

                if (request.id == reply.id and
                    request.sequence == reply.sequence):
                    return reply

        except socket.timeout:
            raise TimeoutExceeded(request.timeout)

        except OSError as err:
            raise ICMPSocketError(str(err))

    def close(self):
        '''
        Close the socket. It cannot be used after this call.

        '''
        if self._socket:
            self._socket.close()
            self._socket = None

    @property
    def is_closed(self):
        '''
        Indicate whether the socket is closed.

        '''
        return self._socket is None


class ICMPv4Socket(ICMPSocket):
    '''
    Socket for sending and receiving ICMPv4 packets.

    :raises SocketPermissionError: If the permissions are insufficient
        to create the socket.

    '''
    def __init__(self):
        super().__init__(ICMPv4Config)

    @property
    def broadcast(self):
        '''
        Indicate whether broadcast support is enabled on the socket.
        Return a boolean.

        .. note::
           To enable broadcast support:
           icmp_socket.broadcast = True

        '''
        return self._socket.broadcast

    @broadcast.setter
    def broadcast(self, allow):
        self._socket.broadcast = allow


class ICMPv6Socket(ICMPSocket):
    '''
    Socket for sending and receiving ICMPv6 packets.

    :raises SocketPermissionError: If the permissions are insufficient
        to create the socket.

    '''
    def __init__(self):
        super().__init__(ICMPv6Config)



'''
    icmplib
    ~~~~~~~

        https://github.com/ValentinBELYN/icmplib

    :copyright: Copyright 2017-2020 Valentin BELYN.
    :license: GNU LGPLv3, see the LICENSE for details.

    ~~~~~~~

    This program is free software: you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public License
    as published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this program.  If not, see
    <https://www.gnu.org/licenses/>.
'''

from .sockets import ICMPv4Socket, ICMPv6Socket
from .models import ICMPRequest, Hop
from .exceptions import *
from .utils import PID, is_ipv6_address

from time import sleep


def traceroute(address, count=3, interval=0.05, timeout=2, id=PID,
        max_hops=30, fast_mode=False):
    '''
    Determine the route to a destination host.

    The Internet is a large and complex aggregation of network hardware,
    connected together by gateways. Tracking the route one's packets
    follow can be difficult. This function utilizes the IP protocol
    time to live field and attempts to elicit an ICMP TIME_EXCEEDED
    response from each gateway along the path to some host.

    :type address: str
    :param address: The destination IP address.

    :type count: int
    :param count: (Optional) The number of ping to perform per hop.

    :type interval: int or float
    :param interval: (Optional) The interval in seconds between sending
        each packet.

    :type timeout: int or float
    :param timeout: (Optional) The maximum waiting time for receiving
        a reply in seconds.

    :type id: int
    :param id: (Optional) The identifier of the request. Used to match
        the reply with the request. In practice, a unique identifier is
        used for every ping process.

    :type max_hops: int
    :param max_hops: (Optional) The maximum time to live (max number of
        hops) used in outgoing probe packets.

    :type fast_mode: bool
    :param fast_mode: (Optional) When this option is enabled and an
        intermediate router has been reached, skip to the next hop
        rather than perform additional requests. The count option then
        becomes the maximum number of requests in case of no responses.

    :rtype: list of Hop
    :returns: A list of Hop objects representing the route to the
        desired host. The list is sorted in ascending order according
        to the distance (in terms of hops) that separates the remote
        host from the current machine.

    :raises SocketPermissionError: If the permissions are insufficient
        to create a socket.

    Usage::

        >>> from icmplib import traceroute
        >>> hops = traceroute('1.1.1.1')
        >>> last_distance = 0

        >>> for hop in hops:
        ...     if last_distance + 1 != hop.distance:
        ...         print('Some routers are not responding')
        ...
        ...     print(f'{hop.distance} {hop.address} {hop.avg_rtt} ms')
        ...
        ...     last_distance = hop.distance
        ...
        1       10.0.0.1             5.19 ms
        2       194.149.169.49       7.55 ms
        3       194.149.166.54       12.2 ms
        *       Some routers are not responding
        5       212.73.205.22        22.1 ms
        6       1.1.1.1              13.5 ms

    See the Hop class for details.

    '''
    if is_ipv6_address(address):
        socket = ICMPv6Socket()

    else:
        socket = ICMPv4Socket()

    ttl = 1
    host_reached = False
    hops = []

    while not host_reached and ttl <= max_hops:
        reply = None
        transmitted_packets = 0
        received_packets = 0

        min_rtt = float('inf')
        avg_rtt = 0.0
        max_rtt = 0.0

        for sequence in range(count):
            request = ICMPRequest(
                destination=address,
                id=id,
                sequence=sequence,
                timeout=timeout,
                ttl=ttl)

            try:
                socket.send(request)
                transmitted_packets += 1

                reply = socket.receive()
                reply.raise_for_status()
                host_reached = True

            except TimeExceeded:
                sleep(interval)

            except ICMPLibError:
                continue

            received_packets += 1

            round_trip_time = (reply.time - request.time) * 1000
            avg_rtt += round_trip_time
            min_rtt = min(round_trip_time, min_rtt)
            max_rtt = max(round_trip_time, max_rtt)

            if fast_mode:
                break

        if reply:
            hop = Hop(
                address=reply.source,
                min_rtt=min_rtt,
                avg_rtt=avg_rtt,
                max_rtt=max_rtt,
                transmitted_packets=transmitted_packets,
                received_packets=received_packets,
                distance=ttl)

            hops.append(hop)

        ttl += 1

    return hops



'''
    icmplib
    ~~~~~~~

        https://github.com/ValentinBELYN/icmplib

    :copyright: Copyright 2017-2020 Valentin BELYN.
    :license: GNU LGPLv3, see the LICENSE for details.

    ~~~~~~~

    This program is free software: you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public License
    as published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this program.  If not, see
    <https://www.gnu.org/licenses/>.
'''

from .sockets import ICMPv4Socket, ICMPv6Socket
from .models import ICMPRequest, Host
from .exceptions import *
from .utils import PID, is_ipv6_address

from threading import Thread
from time import sleep


class PingThread(Thread):
    def __init__(self, **kwargs):
        super().__init__()
        self._kwargs = kwargs
        self._host = None

    def run(self):
        self._host = ping(**self._kwargs)

    @property
    def host(self):
        return self._host


def ping(address, count=4, interval=1, timeout=2, id=PID, **kwargs):
    '''
    Send ICMP ECHO_REQUEST packets to a network host.

    :type address: str
    :param address: The IP address of the gateway or host to which
        the message should be sent.

    :type count: int
    :param count: (Optional) The number of ping to perform.

    :type interval: int or float
    :param interval: (Optional) The interval in seconds between sending
        each packet.

    :type timeout: int or float
    :param timeout: (Optional) The maximum waiting time for receiving
        a reply in seconds.

    :type id: int
    :param id: (Optional) The identifier of the request. Used to match
        the reply with the request. In practice, a unique identifier is
        used for every ping process.

    :param **kwargs: (Optional) Advanced use: arguments passed to the
        ICMPRequest object.

    :rtype: Host
    :returns: A Host object containing statistics about the desired
        destination.

    :raises SocketPermissionError: If the permissions are insufficient
        to create a socket.

    Usage::

        >>> from icmplib import ping
        >>> host = ping('1.1.1.1')
        >>> host.avg_rtt
        13.2
        >>> host.is_alive
        True

    See the Host class for details.

    '''
    if is_ipv6_address(address):
        socket = ICMPv6Socket()

    else:
        socket = ICMPv4Socket()

    transmitted_packets = 0
    received_packets = 0

    min_rtt = float('inf')
    avg_rtt = 0.0
    max_rtt = 0.0

    for sequence in range(count):
        request = ICMPRequest(
            destination=address,
            id=id,
            sequence=sequence,
            timeout=timeout,
            **kwargs)

        try:
            socket.send(request)
            transmitted_packets += 1

            reply = socket.receive()
            reply.raise_for_status()
            received_packets += 1

            round_trip_time = (reply.time - request.time) * 1000
            avg_rtt += round_trip_time
            min_rtt = min(round_trip_time, min_rtt)
            max_rtt = max(round_trip_time, max_rtt)

            if sequence < count - 1:
                sleep(interval)

        except ICMPLibError:
            pass

    if received_packets:
        avg_rtt /= received_packets

    else:
        min_rtt = 0.0

    host = Host(
        address=address,
        min_rtt=min_rtt,
        avg_rtt=avg_rtt,
        max_rtt=max_rtt,
        transmitted_packets=transmitted_packets,
        received_packets=received_packets)

    return host


def multiping(addresses, count=2, interval=1, timeout=2, id=PID,
        max_threads=10, **kwargs):
    '''
    Send ICMP ECHO_REQUEST packets to multiple network hosts.

    :type addresses: list of str
    :param addresses: The IP addresses of the gateways or hosts to
        which messages should be sent.

    :type count: int
    :param count: (Optional) The number of ping to perform per address.

    :type interval: int or float
    :param interval: (Optional) The interval in seconds between sending
        each packet.

    :type timeout: int or float
    :param timeout: (Optional) The maximum waiting time for receiving
        a reply in seconds.

    :type id: int
    :param id: (Optional) The identifier of the requests. This
        identifier will be incremented by one for each destination.

    :type max_threads: int
    :param max_threads: (Optional) The number of threads allowed to
        speed up processing.

    :param **kwargs: (Optional) Advanced use: arguments passed to the
        ICMPRequest object.

    :rtype: list of Host
    :returns: A list of Host objects containing statistics about the
        desired destinations. The list is sorted in the same order as
        the addresses passed in parameters.

    :raises SocketPermissionError: If the permissions are insufficient
        to create a socket.

    Usage::

        >>> from icmplib import multiping
        >>> hosts = multiping(['10.0.0.5', '127.0.0.1', '::1'])

        >>> for host in hosts:
        ...     if host.is_alive:
        ...         print(f'{host.address} is alive!')
        ...
        ...     else:
        ...         print(f'{host.address} is dead!')
        ...
        10.0.0.5 is dead!
        127.0.0.1 is alive!
        ::1 is alive!

    See the Host class for details.

    '''
    hosts = []
    inactive_threads = []
    active_threads = []

    for i, address in enumerate(addresses):
        thread = PingThread(
            address=address,
            count=count,
            interval=interval,
            timeout=timeout,
            id=id + i,
            **kwargs)

        inactive_threads.append(thread)

    while inactive_threads:
        thread = inactive_threads.pop(0)
        thread.start()
        active_threads.append(thread)

        if (inactive_threads and
            len(active_threads) < max_threads):
            sleep(0.05)
            continue

        while active_threads:
            thread = active_threads.pop(0)
            thread.join()
            hosts.append(thread.host)

    return hosts



'''
    icmplib
    ~~~~~~~

        https://github.com/ValentinBELYN/icmplib

    :copyright: Copyright 2017-2020 Valentin BELYN.
    :license: GNU LGPLv3, see the LICENSE for details.

    ~~~~~~~

    This program is free software: you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public License
    as published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this program.  If not, see
    <https://www.gnu.org/licenses/>.
'''

from .exceptions import *
from .utils import is_ipv6_address


class ICMPRequest:
    '''
    A user-created object that represents an ICMP ECHO_REQUEST.

    :type destination: str
    :param destination: The IP address of the gateway or host to which
        the message should be sent.

    :type id: int
    :param id: The identifier of the request. Used to match the reply
        with the request. In practice, a unique identifier is used for
        every ping process.

    :type sequence: int
    :param sequence: The sequence number. Used to match the reply with
        the request. Typically, the sequence number is incremented for
        each packet sent during the process.

    :type payload: bytes
    :param payload: (Optional) The payload content in bytes. Its size
        must be even. A random payload is used by default.

    :type payload_size: int
    :param payload_size: (Optional) The payload size (even number).
        Ignored when the 'payload' parameter is set.

    :type timeout: int or float
    :param timeout: (Optional) The maximum waiting time for receiving
        the reply in seconds.

    :type ttl: int
    :param ttl: (Optional) The time to live of the packet in seconds.

    '''
    def __init__(self, destination, id, sequence, payload=None,
            payload_size=56, timeout=2, ttl=64):

        id &= 0xffff
        sequence &= 0xffff

        if payload:
            payload_size = len(payload)

        self._destination = destination
        self._id = id
        self._sequence = sequence
        self._payload = payload
        self._payload_size = payload_size
        self._timeout = timeout
        self._ttl = ttl
        self._time = 0

    def __repr__(self):
        return f'<ICMPRequest [{self._destination}]>'

    @property
    def destination(self):
        '''
        The IP address of the gateway or host to which the message
        should be sent.

        '''
        return self._destination

    @property
    def id(self):
        '''
        The identifier of the request.
        Used to match the reply with the request.

        '''
        return self._id

    @property
    def sequence(self):
        '''
        The sequence number.
        Used to match the reply with the request.

        '''
        return self._sequence

    @property
    def payload(self):
        '''
        The payload content in bytes.
        Return None if the payload is random.

        '''
        return self._payload

    @property
    def payload_size(self):
        '''
        The payload size.

        '''
        return self._payload_size

    @property
    def timeout(self):
        '''
        The maximum waiting time for receiving the reply in seconds.

        '''
        return self._timeout

    @property
    def ttl(self):
        '''
        The time to live of the packet in seconds.

        '''
        return self._ttl

    @property
    def time(self):
        '''
        The timestamp of the ICMP request.
        Initialized to zero when creating the request and replaced by
        ICMPv4Socket or ICMPv6Socket with the time of sending.

        '''
        return self._time


class ICMPReply:
    '''
    A class that represents an ICMP reply. Generated from an ICMPSocket
    object (ICMPv4Socket or ICMPv6Socket).

    :type source: str
    :param source: The IP address of the gateway or host that composes
        the ICMP message.

    :type id: int
    :param id: The identifier of the reply. Used to match the reply
        with the request.

    :type sequence: int
    :param sequence: The sequence number. Used to match the reply with
        the request.

    :type type: int
    :param type: The type of message.

    :type code: int
    :param code: The error code.

    :type received_bytes: int
    :param received_bytes: The number of bytes received.

    :type time: float
    :param time: The timestamp of the ICMP reply.

    '''
    def __init__(self, source, id, sequence, type, code,
            received_bytes, time):

        self._source = source
        self._id = id
        self._sequence = sequence
        self._type = type
        self._code = code
        self._received_bytes = received_bytes
        self._time = time

    def __repr__(self):
        return f'<ICMPReply [{self._source}]>'

    def raise_for_status(self):
        '''
        Throw an exception if the reply is not an ICMP ECHO_REPLY.
        Otherwise, do nothing.

        :raises ICMPv4DestinationUnreachable: If the ICMPv4 reply is
            type 3.
        :raises ICMPv4TimeExceeded: If the ICMPv4 reply is type 11.
        :raises ICMPv6DestinationUnreachable: If the ICMPv6 reply is
            type 1.
        :raises ICMPv6TimeExceeded: If the ICMPv6 reply is type 3.
        :raises ICMPError: If the reply is of another type and is not
            an ICMP ECHO_REPLY.

        '''
        ipv4_errors = {
            3:  ICMPv4DestinationUnreachable,
            11: ICMPv4TimeExceeded
        }

        ipv6_errors = {
            1:  ICMPv6DestinationUnreachable,
            3:  ICMPv6TimeExceeded
        }

        if is_ipv6_address(self._source):
            expected_type = 129
            errors = ipv6_errors

        else:
            expected_type = 0
            errors = ipv4_errors

        if self._type in errors:
            raise errors[self._type](self)

        if self._type != expected_type:
            message = f'Error type: {self._type}, ' \
                      f'code: {self._code}'

            raise ICMPError(message, self)

    @property
    def source(self):
        '''
        The IP address of the gateway or host that composes the ICMP
        message.

        '''
        return self._source

    @property
    def id(self):
        '''
        The identifier of the request.
        Used to match the reply with the request.

        '''
        return self._id

    @property
    def sequence(self):
        '''
        The sequence number.
        Used to match the reply with the request.

        '''
        return self._sequence

    @property
    def type(self):
        '''
        The type of message.

        '''
        return self._type

    @property
    def code(self):
        '''
        The error code.

        '''
        return self._code

    @property
    def received_bytes(self):
        '''
        The number of bytes received.

        '''
        return self._received_bytes

    @property
    def time(self):
        '''
        The timestamp of the ICMP reply.

        '''
        return self._time


class Host:
    '''
    A class that represents a host. Simplifies the exploitation of
    results from ping and traceroute functions.

    :type address: str
    :param address: The IP address of the gateway or host that
        responded to the request.

    :type min_rtt: float
    :param min_rtt: The minimum round-trip time.

    :type avg_rtt: float
    :param avg_rtt: The average round-trip time.

    :type max_rtt: float
    :param max_rtt: The maximum round-trip time.

    :type transmitted_packets: int
    :param transmitted_packets: The number of packets transmitted to
        the destination host.

    :type received_packets: int
    :param received_packets: The number of packets sent by the remote
        host and received by the current host.

    '''
    def __init__(self, address, min_rtt, avg_rtt, max_rtt,
            transmitted_packets, received_packets):

        self._address = address
        self._min_rtt = round(min_rtt, 3)
        self._avg_rtt = round(avg_rtt, 3)
        self._max_rtt = round(max_rtt, 3)
        self._transmitted_packets = transmitted_packets
        self._received_packets = received_packets

    def __repr__(self):
        return f'<Host [{self._address}]>'

    @property
    def address(self):
        '''
        The IP address of the gateway or host that responded to the
        request.

        '''
        return self._address

    @property
    def min_rtt(self):
        '''
        The minimum round-trip time.

        '''
        return self._min_rtt

    @property
    def avg_rtt(self):
        '''
        The average round-trip time.

        '''
        return self._avg_rtt

    @property
    def max_rtt(self):
        '''
        The maximum round-trip time.

        '''
        return self._max_rtt

    @property
    def transmitted_packets(self):
        '''
        The number of packets transmitted to the destination host.

        '''
        return self._transmitted_packets

    @property
    def received_packets(self):
        '''
        The number of packets sent by the remote host and received by
        the current host.

        '''
        return self._received_packets

    @property
    def packet_loss(self):
        '''
        Packet loss occurs when packets fail to reach their destination.
        Return a float between 0 and 1 (all packets are lost).

        '''
        if not self._transmitted_packets:
            return 0.0

        return (1 - self._received_packets
                  / self._transmitted_packets)

    @property
    def is_alive(self):
        '''
        Return True if the host is reachable, False otherwise.

        '''
        return self.packet_loss < 1


class Hop(Host):
    '''
    A class that represents a hop. It extends the Host class and adds
    some features for traceroute.

    :type address: str
    :param address: The IP address of the gateway or host that
        responded to the request.

    :type min_rtt: float
    :param min_rtt: The minimum round-trip time.

    :type avg_rtt: float
    :param avg_rtt: The average round-trip time.

    :type max_rtt: float
    :param max_rtt: The maximum round-trip time.

    :type transmitted_packets: int
    :param transmitted_packets: The number of packets transmitted to
        the destination host.

    :type received_packets: int
    :param received_packets: The number of packets sent by the remote
        host and received by the current host.

    :type distance: int
    :param distance: The distance (in terms of hops) that separates the
        remote host from the current machine.

    '''
    def __init__(self, address, min_rtt, avg_rtt, max_rtt,
            transmitted_packets, received_packets, distance):

        super().__init__(address, min_rtt, avg_rtt, max_rtt,
            transmitted_packets, received_packets)

        self._distance = distance

    def __repr__(self):
        return f'<Hop {self._distance} [{self._address}]>'

    @property
    def distance(self):
        '''
        The distance (in terms of hops) that separates the remote host
        from the current machine.

        '''
        return self._distance



'''
    icmplib
    ~~~~~~~

        https://github.com/ValentinBELYN/icmplib

    :copyright: Copyright 2017-2020 Valentin BELYN.
    :license: GNU LGPLv3, see the LICENSE for details.

    ~~~~~~~

    This program is free software: you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public License
    as published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this program.  If not, see
    <https://www.gnu.org/licenses/>.
'''

import socket


# Fix for Windows
if not hasattr(socket, 'IPPROTO_IPV6'):
    socket.IPPROTO_IPV6 = 41


class IPSocket:
    def __init__(self, family, protocol):
        self._socket = socket.socket(
            family=family,
            type=socket.SOCK_RAW,
            proto=protocol)

        self.timeout = 5
        self.ttl = 64

    def send(self, packet, address, port):
        return self._socket.sendto(packet, (address, port))

    def receive(self, buffer_size=1024):
        result = self._socket.recvfrom(buffer_size)

        packet = result[0]
        address = result[1][0]
        port = result[1][1]

        return packet, address, port

    def close(self):
        self._socket.close()

    @property
    def timeout(self):
        return self._timeout

    @timeout.setter
    def timeout(self, timeout):
        self._socket.settimeout(timeout)
        self._timeout = timeout

    @property
    def ttl(self):
        return self._ttl

    @ttl.setter
    def ttl(self, ttl):
        self._ttl = ttl


class IPv4Socket(IPSocket):
    def __init__(self, protocol):
        super().__init__(
            family=socket.AF_INET,
            protocol=protocol)

        self._broadcast = False

    @IPSocket.ttl.setter
    def ttl(self, ttl):
        self._ttl = ttl

        self._socket.setsockopt(
            socket.SOL_IP,
            socket.IP_TTL,
            ttl)

    @property
    def broadcast(self):
        return self._broadcast

    @broadcast.setter
    def broadcast(self, allow):
        self._broadcast = allow

        self._socket.setsockopt(
            socket.SOL_SOCKET,
            socket.SO_BROADCAST,
            allow)


class IPv6Socket(IPSocket):
    def __init__(self, protocol):
        super().__init__(
            family=socket.AF_INET6,
            protocol=protocol)

    @IPSocket.ttl.setter
    def ttl(self, ttl):
        self._ttl = ttl

        self._socket.setsockopt(
            socket.IPPROTO_IPV6,
            socket.IPV6_MULTICAST_HOPS,
            ttl)



'''
    icmplib
    ~~~~~~~

        https://github.com/ValentinBELYN/icmplib

    :copyright: Copyright 2017-2020 Valentin BELYN.
    :license: GNU LGPLv3, see the LICENSE for details.

    ~~~~~~~

    This program is free software: you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public License
    as published by the Free Software Foundation, either version 3 of
    the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this program.  If not, see
    <https://www.gnu.org/licenses/>.
'''


class ICMPLibError(Exception):
    '''
    Exception class for the icmplib package.

    '''
    def __init__(self, message):
        self._message = message

    def __str__(self):
        return self._message

    @property
    def message(self):
        return self._message


class ICMPSocketError(ICMPLibError):
    '''
    Base class for ICMP sockets exceptions.

    '''


class SocketPermissionError(ICMPSocketError):
    '''
    Raised when the permissions are insufficient to create a socket.

    '''
    def __init__(self):
        message = 'Root privileges are required to create the socket'
        super().__init__(message)


class SocketBroadcastError(ICMPSocketError):
    '''
    Raised when a broadcast address is used and the corresponding
    option is not enabled on the socket.

    '''
    def __init__(self):
        message = 'Broadcast is not allowed: ' \
                  'please use broadcast method (setter) to allow it'
        super().__init__(message)


class TimeoutExceeded(ICMPSocketError):
    '''
    Raised when a timeout occurs on a socket.

    '''
    def __init__(self, timeout):
        message = f'The timeout has been reached ({timeout}s)'
        super().__init__(message)


class ICMPError(ICMPLibError):
    '''
    Base class for ICMP error messages.

    '''
    def __init__(self, message, reply):
        super().__init__(message)
        self._reply = reply

    @property
    def reply(self):
        return self._reply


class DestinationUnreachable(ICMPError):
    '''
    Base class for ICMP Destination Unreachable messages.

    Destination Unreachable message is generated by the host or its
    inbound gateway to inform the client that the destination is
    unreachable for some reason.

    '''
    _CODES = {}

    def __init__(self, reply):
        if reply.code in self._CODES:
            message = self._CODES[reply.code]

        else:
            message = f'Destination unreachable, bad code: {reply.code}'

        super().__init__(message, reply)


class ICMPv4DestinationUnreachable(DestinationUnreachable):
    _CODES = {
        0:  'Destination network unreachable',
        1:  'Destination host unreachable',
        2:  'Destination protocol unreachable',
        3:  'Destination port unreachable',
        4:  'Fragmentation needed and DF set',
        5:  'Source route failed',
        6:  'Destination network unknown',
        7:  'Destination host unknown',
        8:  'Source host isolated',
        9:  'Destination network prohibed',
        10: 'Destination host prohibed',
        11: 'Destination network unreachable for ToS',
        12: 'Destination host unreachable for ToS',
        13: 'Packet filtered',
        14: 'Precedence violation',
        15: 'Precedence cutoff'
    }


class ICMPv6DestinationUnreachable(DestinationUnreachable):
    _CODES = {
        0:  'No route to destination',
        1:  'Communication with destination administratively prohibited',
        2:  'Beyond scope of source address',
        3:  'Address unreachable',
        4:  'Port unreachable',
        5:  'Source address failed ingress/egress policy',
        6:  'Reject route to destination'
    }


class TimeExceeded(ICMPError):
    '''
    Base class for ICMP Time Exceeded messages.

    Time Exceeded message is generated by a gateway to inform the
    source of a discarded datagram due to the time to live field
    reaching zero. A Time Exceeded message may also be sent by a host
    if it fails to reassemble a fragmented datagram within its time
    limit.

    '''
    _CODES = {}

    def __init__(self, reply):
        if reply.code in self._CODES:
            message = self._CODES[reply.code]

        else:
            message = f'Time exceeded, bad code: {reply.code}'

        super().__init__(message, reply)


class ICMPv4TimeExceeded(TimeExceeded):
    _CODES = {
        0:  'Time to live exceeded',
        1:  'Fragment reassembly time exceeded'
    }


class ICMPv6TimeExceeded(TimeExceeded):
    _CODES = {
        0:  'Hop limit exceeded',
        1:  'Fragment reassembly time exceeded'
    }
